//
// Copyright 2003 Sony Corporation 
//
// Permission to use, copy, modify, and redistribute this software for
// non-commercial use is hereby granted.
//
// This software is provided "as is" without warranty of any kind,
// either expressed or implied, including but not limited to the
// implied warranties of fitness for a particular purpose.
//

#include <OPENR/OSyslog.h>
#include <OPENR/OFbkImage.h>
#include "JPEGEncoder.h"
#include "write_jpeg.h"

JPEGEncoder::JPEGEncoder() : ipstackRef(),
                             pixelInterleavedYCbCr(0),
                             freeJpegBufList()
{
}

bool
JPEGEncoder::Init(const antStackRef& ipstack)
{
    ipstackRef = ipstack;

    pixelInterleavedYCbCr = (byte*)malloc(IMAGE_BUFSIZE);
    if (pixelInterleavedYCbCr == 0) {
        OSYSLOG1((osyslogERROR, "JPEGEncoder::Init() : NO MEMORY"));
        return false;
    }

    for (int i = 0; i < NUM_JPEG_BUF; i++) {

        antEnvCreateSharedBufferMsg jpegBufferMsg(JPEG_BUFSIZE);
        
        jpegBufferMsg.Call(ipstackRef, sizeof(jpegBufferMsg));
        if (jpegBufferMsg.error != ANT_SUCCESS) {
            OSYSLOG1((osyslogERROR, "%s : NO MEMORY antError %d",
                      "JPEGEncoder::Init()", jpegBufferMsg.error));
            return false;
        }

        jpegBuf[i] = jpegBufferMsg.buffer;
        jpegBuf[i].Map();
        freeJpegBufList.push_back((byte*)jpegBuf[i].GetAddress());
    }

    return true;
}

void
JPEGEncoder::FreeJPEG(byte* jpeg)
{
    Free(jpeg);
}

bool
JPEGEncoder::GetJPEG(OFbkImageVectorData* imageVec,
                     OFbkImageLayer layer, int quality, byte** jpeg, int* size)
{
    OFbkImageInfo* info = imageVec->GetInfo(layer);
    byte*          data = imageVec->GetData(layer);

    OFbkImage yImg(info, data, ofbkimageBAND_Y);
    OFbkImage crImg(info, data, ofbkimageBAND_Cr);
    OFbkImage cbImg(info, data, ofbkimageBAND_Cb);

    int w = yImg.Width();
    int h = yImg.Height();
    int skip = yImg.Skip();

    byte* iptr = (byte*)pixelInterleavedYCbCr;

    for (int y = 0; y < h; y++) {

        byte* yline  = yImg.Pointer()  + (w + skip) * y;
        byte* crline = crImg.Pointer() + (w + skip) * y;
        byte* cbline = cbImg.Pointer() + (w + skip) * y;

        for (int x = 0; x < w; x++) {
            *iptr++ = yline[x];     // Y
            *iptr++ = cbline[x];    // Cb
            *iptr++ = crline[x];    // Cr
        }

    }

    byte* jbuf = Allocate();
    if (jbuf == 0) return false;

    int jsize = write_jpeg_mem(pixelInterleavedYCbCr,
                               w, h, quality, jbuf, 3*w*h);
    *jpeg = jbuf;
    *size = jsize;

    return true;
}

bool
JPEGEncoder::GetJPEG(byte *yImage, int quality, byte** jpeg, int* size)
{
	byte* iptr = (byte*)pixelInterleavedYCbCr;
	int h = 144;
	int w = 176;
	for (int y = 0; y < h; y++) 
		for (int x = 0; x < w; x++) {
			*iptr++ = yImage[y*w+x];     // Y
			*iptr++ = 128;    // Cb
			*iptr++ = 128;    // Cr
		}

	byte* jbuf = Allocate();
	if (jbuf == 0) return false;

	int jsize = write_jpeg_mem(pixelInterleavedYCbCr,
			w, h, quality, jbuf, 3*w*h);
	*jpeg = jbuf;
	*size = jsize;

	return true;
}


void
JPEGEncoder::Save(OFbkImageVectorData* imageVec,
		OFbkImageLayer layer, int quality, char* path)
{
	OFbkImageInfo* info = imageVec->GetInfo(layer);
	byte*          data = imageVec->GetData(layer);

	OFbkImage yImg(info, data, ofbkimageBAND_Y);
	OFbkImage crImg(info, data, ofbkimageBAND_Cr);
	OFbkImage cbImg(info, data, ofbkimageBAND_Cb);

	int w = yImg.Width();
	int h = yImg.Height();
	int skip = yImg.Skip();

	byte* iptr = (byte*)pixelInterleavedYCbCr;

	for (int y = 0; y < h; y++) {

		byte* yline  = yImg.Pointer()  + (w + skip) * y;
		byte* crline = crImg.Pointer() + (w + skip) * y;
		byte* cbline = cbImg.Pointer() + (w + skip) * y;

		for (int x = 0; x < w; x++) {
			*iptr++ = yline[x];     // Y
			*iptr++ = cbline[x];    // Cb
			*iptr++ = crline[x];    // Cr
		}

	}

	FILE* fp = fopen(path, "w");
	if (fp == NULL) {
		OSYSLOG1((osyslogERROR, "JPEGEncoder::Save() : can't open %s", path));
		return;
	}
	write_jpeg_file(pixelInterleavedYCbCr, w, h, quality, fp);
	fclose(fp);
}

	byte*
JPEGEncoder::Allocate()
{
	if (freeJpegBufList.size() == 0) return 0;

	byte* buf = freeJpegBufList.front();
	freeJpegBufList.pop_front();
	return buf;
}

	void
JPEGEncoder::Free(byte* buf)
{
	freeJpegBufList.push_back(buf);
}
